"""
Module for handling errors and presenting them as dialogs.
"""

import __builtin__
import sys
import traceback
from ErrorFormatter import ErrorFormatter
import dialog


class UserError(StandardError):
    """
    Exception for user-visible errors.
    """
    
    def __init__(self, error, description,
                 code = None, lineno = -1, show_details = True):
        
        self.error = error
        self.description = description
        if (code): self.code = code
        if (lineno != -1): self.lineno = lineno
        self.show_details = show_details
        
        StandardError.__init__(self)



class _Error:

    def __init__(self):

        # table for mapping IDs to the .display files
        self.__ident_map = {"unknown": ""}

        # table for mapping filenames to code pieces
        self.__code_map = {}

        

    def register(self, ident, path):
        """
        Registers the given ID with the given path. This is needed so that
        the dialog can point to the broken desklet.
        """
    
        self.__ident_map[ident] = path



    def forget(self, ident):
        """
        Forgets about the given ID.
        """

        del self.__ident_map[ident]
        dialog.forget(ident)



    def register_code(self, filename, code):
        """
        Registers a piece of code with a "filename". This is needed for code
        which is not accessible as a file. The filename is a pseudo-filename,
        not a real one.
        """

        self.__code_map[filename] = code



    def __dig_down(self, tb):
        """
        Digs through the traceback looking for the place which is interesting
        for the exception, which is not necessarily where it was thrown.
        Returns a tuple (filename, lineno).
        """

        tbs = traceback.extract_tb(tb)
        filename = None
        lineno = -1
        
        found = False
        while (True and tbs):
            trace =  tbs.pop()
            filename = trace[0]
            lineno = trace[1]
            if (filename.startswith("<")):
                found = True
                break
            
        del tbs

        if (found):
            return (filename, lineno)
        else:
            return (None, -1)
    


    def handle(self, ident):
        """
        Handles the recently caught error and associates it with the given
        display ID.
        """

        exc_type, exc_value, tb = sys.exc_info()
        dspfile = self.__ident_map[ident]

        filename = None
        lineno = -1
        code = None

        # get some information from the exception
        if (hasattr(exc_value, "lineno")): lineno = exc_value.lineno
        if (hasattr(exc_value, "filename")): filename = exc_value.filename
        if (hasattr(exc_value, "code")): code = exc_value.code
        if (hasattr(exc_value, "show_details")):
            show_details = exc_value.show_details
        else:
            show_details = True


        # find the interesting piece of code
        if (not filename and lineno == -1):
            filename, lineno = self.__dig_down(tb)
        del tb

            
        if (not code): code = self.__code_map.get(filename)

        if (filename):
            if (filename.startswith("<")):
                index1 = filename.find("'")
                index2 = filename.rfind("_")
                filename = filename[index1 + 1:index2]


        # hilight bad code
        if (show_details):
            details = ErrorFormatter().format(sys.exc_info(), code, lineno,
                                              filename = filename)
        else:
            details = ""


        # present exception to the user
        if (hasattr(exc_value, "error") and hasattr(exc_value, "description")):
            log("=== " + exc_value.error + "\n" + details)
            dialog.user_error(ident,
                              exc_value.error +
                              "\n<i><small>%s</small></i>\n\n" % dspfile,
                              exc_value.description, details)
        else:
            log("=== Runtime Error\n" + details)
            dialog.user_error(ident,
                              _("Runtime Error") +
                              "\n<i><small>%s</small></i>\n\n" % dspfile,
                              _("An error occured while executing a desklet."),
                              details)




# make UserError globally available
__builtin__.UserError = UserError

_singleton = _Error()
def Error(): return _singleton



from ConfigWidget import ConfigWidget
from utils.datatypes import *

import gtk

class ConfigString(ConfigWidget):

    def __init__(self, name, getter, setter, caller):

        ConfigWidget.__init__(self, name, getter, setter, caller)

        self._register_property("password", TYPE_BOOL, self._setp_password,
                                self._getp, False,
                                doc = "Whether to obfuscate input")
        self._register_property("value", TYPE_STRING, self._setp_value,
                                self._getp, "", doc = "Value")




    def get_widgets(self):

        self.__label = gtk.Label("")
        self.__label.show()
        self.__entry = gtk.Entry()
        self.__entry.set_width_chars(15)
        self.__entry.show()
        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.show()
        align.add(self.__label)

        self.__entry.set_invisible_char(unichr(0x2022))
        self.__entry.connect("focus-out-event", self.__on_change)

        return (align, self.__entry)


    def __on_change(self, src, event):

        value = src.get_text()
        self._set_config(value)


    def _set_enabled(self, value): self.__entry.set_sensitive(value)
    def _set_label(self, value): self.__label.set_text(value)


    def _setp_password(self, key, value):

        self.__entry.set_visibility(not value)
        self._setp(key, value)


    def _setp_value(self, key, value):

        self.__entry.set_text(value)
        self._set_config(value)
        self._setp(key, value)



from DisplayTarget import DisplayTarget
from utils.datatypes import *

import gtk


#
# Class for a target that embeds Bonobo controls.
#
class TargetBonoboControl(DisplayTarget):

    def __init__(self, name, parent):

        # the control; you can load a control only once
        self.__control = None

        self.__widget = gtk.HBox()
        self.__widget.show()

        DisplayTarget.__init__(self, name, parent)

        self._register_property("oafiid", TYPE_STRING,
                                self._setp_oafiid, self._getp)


    def get_widget(self): return self.__widget


    def _setp_oafiid(self, key, value):

        import bonobo.ui
        try:
            container = bonobo.ui.Container()
            control = bonobo.ui.Widget(str(value),
                                           container.corba_objref())
            pbag = control.get_control_frame().get_control_property_bag()
            slots = pbag.getKeys("")

            control.show()

            if self.__control:  # we have to remove the previous control
                self.remove( self.__control )

            self.__widget.add(control)
            self.__control = control
            self._setp(key, value)

        except StandardError, exc:
            log("Warning: An error occured while setting the oafiid:\n%s" \
                % (exc,))




from libdesklets.controls import Interface, Permission

class ITime(Interface):

    # properties along with their permissions
    time = Permission.READ
    date = Permission.READ
    timezone = Permission.READWRITE



from libdesklets.system.Arch import Arch

import re
import struct
import os

class Generic(Arch):

    def __init__(self):

        Arch.__init__(self)

        self.__net_devices = re.compile('^\s*(\w+):.*mtu', re.M
                                        ).findall( os.popen('/sbin/ifconfig -a').read() )



    def net_devices(self):
        """
        @return : all available network devices
        @rtype  : list
        """

        return (self.__net_devices)



    def users(self):
        """
        @return : number of connected users
        @rtype  : int
        """

        # man utmp
        # don't know if
        # sizeof(struct utmp) == 44
        # offsetof(struct utmp, ut_name) == 8
        # on every FreeBSD arch

        count = 0
        data = open('/var/run/utmp', 'rb').read()

        for i in range(8, len(data), 44):

            ut_name = struct.unpack('c', data[i:i+1])[0]

            if (ut_name != '\0'):
                count += 1

        return count



from shell.Plugin import Plugin
from main import USERHOME
from utils import dialog
from Package import Package
from Downloader import Downloader

import gobject
import gtk
import os
import tempfile
import thread


#
# Installer for desklet packages.
#
class Installer_Package(Plugin):

    def init(self):

        self.__downloader = Downloader(self._get_path())

        menu = self._get_plugin("UI_Menu")
        menu.set_separator("File/Slot")
        menu.set_item("File/Slot", gtk.STOCK_OPEN,
                      _("I_nstall remote package..."), self.__on_remote_file)
        menu.set_item("File/Slot", gtk.STOCK_OPEN, _("_Install package..."),
                      self.__on_file_open)




    #
    # Installs from the given URI.
    #
    def install_from(self, path):

        thread.start_new_thread(self.__install_file, (path,))


    def __on_file_open(self):

        def f(fsel, self):
            file = fsel.get_filename()
            fsel.destroy()
            self.install_from(file)

        dialog.fileselector("", f, None, self)


    def __on_remote_file(self):

        def responder(src, response):
            text = src.vbox.get_children()[0].get_children()[-1].get_text()
            if (response == gtk.RESPONSE_ACCEPT):
                if (text):
                    self.install_from(text)
            else:
                src.destroy()

        from utils.HIGDialog import HIGDialog
        dialog = HIGDialog(buttons = (gtk.STOCK_CANCEL, gtk.RESPONSE_REJECT,
                                      gtk.STOCK_OK, gtk.RESPONSE_ACCEPT))
        dialog.connect("response", responder)
        dialog.show()
        label = gtk.Label("Please enter a valid URI here")
        label.show()
        entry = gtk.Entry()
        entry.show()

        vbox = dialog.vbox
        hbox = gtk.HBox()
        hbox.show()
        hbox.set_property("spacing", 12)
        hbox.set_property("border-width", 6)
        hbox.pack_start(label)
        hbox.pack_start(entry)
        vbox.pack_start(hbox)


    #
    # Installs the given file.
    #
    def __install_file(self, url):

        if (url.startswith("/")):
            path = url
            is_local = True
        else:
            path = tempfile.mktemp()
            self.__downloader.download(url, path)
            is_local = False

        try:
            pkg = Package(path)
        except IOError:
            gobject.timeout_add(0, dialog.warning, _("Invalid package"),
                                _("The supplied file could not be opened as a "
                                  "package. It is either corrupted or not of "
                                  "the correct file type."))
            return

        display_paths = self.__install_displays(pkg)
        control_paths = self.__install_controls(pkg)
        sensor_paths = pkg.find_sensors()
        pkg.close()
        if (not is_local): os.unlink(path)

        query = ""
        if (display_paths):
            query = "(MATCH 'path' '*/%s/*')" \
                    % (os.path.basename(display_paths[0]))

        if (not display_paths and not control_paths and not sensor_paths):
            gobject.timeout_add(0, dialog.info, _("Installation failed"),
                                _("The package could not be installed because "
                                  "it contained no installable files."))
        else:
            gobject.timeout_add(0, dialog.info, _("Installation complete"),
                                "The package has been installed successfully.")

        collection = self._get_plugin("Core_DisplayCollection")
        displaybrowser = self._get_plugin("UI_DisplayBrowser")
        gobject.timeout_add(0, collection.reload)
        gobject.timeout_add(0, displaybrowser.set_location, query)



    #
    # Installs the displays from the given package.
    #
    def __install_displays(self, pkg):

        dirs = pkg.find_displays()
        for d in dirs:
            os.system("cp -r %s %s/" % (d, os.path.join(USERHOME, "Displays")))
        sensors = pkg.find_sensors()
        for s in sensors:
            os.system("python %s --nomsg" % (s))

        return dirs


    #
    # Installs the controls from the given package.
    #
    def __install_controls(self, pkg):

        import time
        name = str(time.time())
        dirs = pkg.find_controls()
        path = os.path.join(USERHOME, "Controls")
        if (not os.path.exists(path)): os.mkdir(path)
        for d in dirs:
            os.system("cp -r %s %s/"
                      % (d, os.path.join(path, name)))

        return dirs



def get_class(): return Installer_Package



from Device import Device

import threading
import time

class Swap(Device):

    def __init__(self):

        Device.__init__(self)

        self.__pagein  = 0
        self.__pageout = 0
        self.__time    = time.time()
        self.__lock    = threading.Lock()


    def poll(self):

        self.__lock.acquire()

        try:
            swap = gtop.swap()

            now       = time.time()
            page_in   = swap.pagein
            page_out  = swap.pageout

            time_diff = now - self.__time
            in_diff   = pagein  - self.__pagein
            out_diff  = pageout - self.__pageout

            speed_in  = int(in_diff / time_diff)
            speed_out = int(out_diff / time_diff)

            self.__pagein = page_in
            self.__pageout = page_out
            self.__time = now

            return (speed_in, speed_out)

        finally:
            self.__lock.release()




from shell.Plugin import Plugin
from TipOfTheDay import TipOfTheDay
from main import HOME
from main.AboutDialog import AboutDialog

import gtk
import os

UNDEF = "-- undef --"


#
# Plugin for providing the user with help.
#
class Shell_Help(Plugin):

    def init(self):
        def obs_tip(src, cmd, value):
            if (cmd == src.OBS_TIP):
                self._set_config("current_tip", value)
            elif (cmd == src.OBS_TOGGLED):
                self._set_config("show_tip_of_the_day", value)


        current_tip = self._get_config("current_tip")
        if (current_tip == UNDEF): current_tip = 0
        self.__tip = TipOfTheDay(current_tip)
        show_tip = self._get_config("show_tip_of_the_day")
        if (show_tip == UNDEF): show_tip = True
        self.__tip.set_show(show_tip)
        self.__tip.add_observer(obs_tip)
        self.__tip.next_tip()
        if (show_tip): self.__show_tip()

        menu = self._get_plugin("UI_Menu")
        menu.insert("Slot2", "Help")
        menu.set_item("Help", None, _("_Help"), None)
        menu.set_item("Help/Tip", gtk.STOCK_DIALOG_INFO, _("_Tip of the Day"),
                      self.__show_tip)
        menu.set_separator("Help/Separator2")
        menu.set_item("Help/About", "gnome-stock-about", _("_About..."),
                      self.__show_about)

        self.__about = AboutDialog(os.path.join(HOME, "data"))


    def __show_tip(self):

        self.__tip.show()


    def __show_about(self):

        self.__about.show()

        
def get_class(): return Shell_Help



from main import ICON
from utils.systray import Icon as SysTrayIcon
from utils.tiling import Tiling

import gtk


class TrayIcon:

    """ Class for a simple tray icon container. """

    def __init__(self):

        icon = Tiling()
        self.__ebox = gtk.EventBox()
        self.__menu = gtk.Menu()
        self.__trayicon = SysTrayIcon("gDesklets")

        icon.set_from_file(ICON);
        icon.render(24, 24, 1, 1);

        self.__ebox.add(icon)
        self.__ebox.show_all()
        self.__trayicon.add(self.__ebox)

        self.__ebox.connect("button-press-event", self.__on_button)


    def __on_button(self, src, event):

        if (event.button == 3):
            self.__menu.popup(None, None, None, event.button, event.time)


    def set_menu(self, items):

        for entry in items:
            if (entry):
                icon, label, callback = entry
                if (icon):
                    item = gtk.ImageMenuItem(icon)
                    if (label):
                        item.get_children()[0].set_text_with_mnemonic(label)
                else:
                    item = gtk.MenuItem(label)

                if (callback):
                    item.connect("activate", callback)
            else:
                item = gtk.MenuItem()

            item.show()
            self.__menu.append(item)


    def connect(self, event, *args):

        self.__ebox.connect(event, *args)



import threading


def locked(f, lock):

    def helper(*args, **kw):

        lock.acquire()

        try:
            return f(*args, **kw)
        finally:
            lock.release()

    return helper


class SyncWriter(object):

    __slots__ = ('__stream', 'write', 'writelines', 'flush')

    def __init__(self, stream):

        self.__stream = stream

        lock = threading.RLock()

        self.write      = locked(stream.write,      lock)
        self.writelines = locked(stream.writelines, lock)
        self.flush      = locked(stream.flush,      lock)


    def __getattr__(self, name):

        return getattr(self.__stream, name)




from ElementWrapper import ElementWrapper
from ControlWrapper import ControlWrapper
from config.StateSaver import StateSaver, DefaultStateSaver
from Scriptlet import Scriptlet
from display.MenuItem import MenuItem
from utils.ErrorFormatter import ErrorFormatter
from utils import dialog
from layout import Unit
import exceptions


_AUTHORIZED_COMMANDS_KEY = "authorized_commands"

#
# Class for inline scripts together with their environment.
#
class Script:

    def __init__(self, dsp_id, dsp_path):

        self.__scriptlets = {}

        # ID of the display
        self.__dsp_id = dsp_id

        # path of the display
        self.__dsp_path = dsp_path

        # the state saver
        self.__state_saver = StateSaver(dsp_id)

        # the environment for this script
        self.__environment = {}

        # the list of loaded controls
        self.__loaded_controls = []

        # flag indicating whether the display has been stopped
        self.__is_stopped = False


        #
        # setup a sandbox environment
        #
        self.__environment["__builtins__"] = None
        self.__environment["__name__"] = "inline"  # required for classes

        # unit constructor and units
        self.__environment["Unit"] = Unit.Unit
        self.__environment["PX"] = Unit.UNIT_PX
        self.__environment["CM"] = Unit.UNIT_CM
        self.__environment["IN"] = Unit.UNIT_IN
        self.__environment["PT"] = Unit.UNIT_PT
        self.__environment["PERCENT"] = Unit.UNIT_PERCENT

        self.__environment["MenuItem"] = MenuItem

        self.__environment["add_timer"] = self.__script_add_timer
        self.__environment["get_config"] = self.__script_get_config
        self.__environment["get_control"] = self.__script_get_control
        self.__environment["set_config"] = self.__script_set_config
        self.__environment["launch"] = self.__script_launch

        self.__environment["Null"] = Null

        # see end of file
        # WTF do we need True/False there? (because they're not yet keywords)
        self.__environment["False"] = False
        self.__environment["True"] = True
        self.__environment["abs"] = abs
        self.__environment["bool"] = bool
        self.__environment["callable"] = callable
        self.__environment["chr"] = chr
        self.__environment["classmethod"] = classmethod
        self.__environment["cmp"] = cmp
        self.__environment["complex"] = complex
        self.__environment["delattr"] = delattr
        self.__environment["dict"] = dict
        self.__environment["divmod"] = divmod
        self.__environment["enumerate"] = enumerate
        self.__environment["float"] = float
        self.__environment["getattr"] = getattr
        self.__environment["hasattr"] = hasattr
        self.__environment["hash"] = hash
        self.__environment["hex"] = hex
        self.__environment["id"] = id
        self.__environment["int"] = int
        self.__environment["isinstance"] = isinstance
        self.__environment["issubclass"] = issubclass
        self.__environment["iter"] = iter
        self.__environment["len"] = len
        self.__environment["list"] = list
        self.__environment["locals"] = locals
        self.__environment["long"] = long
        self.__environment["max"] = max
        self.__environment["min"] = min
        self.__environment["object"] = object
        self.__environment["oct"] = oct
        self.__environment["ord"] = ord
        self.__environment["property"] = property
        self.__environment["range"] = range
        self.__environment["reduce"] = reduce
        self.__environment["repr"] = repr
        self.__environment["round"] = round
        self.__environment["setattr"] = setattr
        self.__environment["staticmethod"] = staticmethod
        self.__environment["str"] = str
        self.__environment["sum"] = sum
        self.__environment["super"] = super
        self.__environment["tuple"] = tuple
        self.__environment["type"] = type
        self.__environment["unichr"] = unichr
        self.__environment["unicode"] = unicode
        self.__environment["vars"] = vars
        self.__environment["xrange"] = xrange
        self.__environment["zip"] = zip

        # exceptions, we need the exceptions
        for name in dir(exceptions):
            if (not name.startswith("_")):
                exc = getattr(exceptions, name)
                self.__environment[name] = exc
        #end for


    #
    # Handles errors in the script.
    #
    def __handle_error(self):

        from utils.error import Error
        Error().handle(self.__dsp_id)


    #
    # Runs a timer in the sandbox. The timer stops when the script is being
    # stopped.
    #
    def __script_add_timer(self, interval, callback, *args):

        def f():
            try:
                if (self.__is_stopped):
                    return False
                else:
                    ret = callback(*args)
                    return ret

            except:
                self.__handle_error()
                return False


        if (type(interval) not in (type(1), type(1.0), type(1L))
              or interval < 0):
            raise UserError(_("Error in add_timer function"),
                           _("\"%s\" isn't a valid integer value!")
                           % `interval`)
            #dialog.warning(_("Error in add_timer function"),
            #               _("\"%s\" isn't a valid integer value!")
            #               % `interval`)
            #return

        import gobject
        gobject.timeout_add(interval, f)



    #
    # Retrieves a configuration value.
    #
    def __script_get_config(self, key, default = None):

        return self.__state_saver.get_key(key, default)



    #
    # Stores a configuration value.
    #
    def __script_set_config(self, key, value):

        self.__state_saver.set_key(key, value)



    #
    # Returns a control readily wrapped for the sandbox.
    #
    def __script_get_control(self, interface):

        # FIXME: ensure that this does not break the sandbox
        from factory.ControlFactory import ControlFactory
        factory = ControlFactory()
        ctrl = factory.get_control(interface)
        if (ctrl):
            self.__loaded_controls.append(ctrl)
            wrapped = ControlWrapper(ctrl)
            return wrapped

        raise UserError(_("No Control could be found for interface %s") % \
                                                                 (interface,),
                       _("This means that a functionality won't be available "
                         "during execution!"))
        #dialog.warning(_("No Control could be found for interface %s") % \
        #                                                         (interface,),
        #               _("This means that a functionality won't be available "
        #                 "during execution!"))

        #return ""


    #
    # Launches the given command if it's safe.
    #
    def __script_launch(self, command):

        states = DefaultStateSaver()
        permissions = states.get_key(_AUTHORIZED_COMMANDS_KEY, {})

        def run_cmd():
            import os
            os.system(command + " &")

        def run_and_permit():
            permissions[(self.__dsp_id, command)] = True
            states.set_key(_AUTHORIZED_COMMANDS_KEY, permissions)
            run_cmd()


        if ((self.__dsp_id, command) in permissions):
            run_cmd()

        else:
            dialog.question(None,
                            _("Security Risk"),
                            _("The desklet %s wants to execute a system "
                              "command:\n"
                              "\n"
                              "     <tt><b>%s</b></tt>\n"
                              "\n"
                              "To protect your system from malicious "
                              "programs, you can deny the execution of this "
                              "command.\n"
                              "\n"
                              "If you are sure that the command is harmless, "
                              "you may permanently allow this desklet "
                              "instance to run it.")
                            % (self.__dsp_path, command),
                            (_("Deny!"), None),
                            (_("Allow once"), run_cmd),
                            (_("Allow for this desklet"), run_and_permit))


    #
    # Stops this scripting object.
    #
    def stop(self):

        self.__is_stopped = True
        del self.__environment
        for c in self.__loaded_controls:
            try:
                c.stop()
            except StandardError, exc:
                import traceback; traceback.print_exc()
                log("Could not stop control %s" % c)
                del c
        del self.__loaded_controls



    #
    # Removes this scripting object and its state.
    #
    def remove(self):

        states = DefaultStateSaver()
        permissions = states.get_key(_AUTHORIZED_COMMANDS_KEY, {})
        for ident, cmd in permissions.keys():
            if (ident == self.__dsp_id): del permissions[(ident, cmd)]
        states.set_key(_AUTHORIZED_COMMANDS_KEY, permissions)

        self.__state_saver.remove()


    #
    # Creates the given namespace if it does not yet exist.
    #
    def __make_namespace(self, namespace):

        class Namespace: pass
        if (not namespace in self.__environment):
            self.__environment[namespace] = Namespace()


    #
    # Adds the given element to the environment. The namespace has to exist.
    # If no namespace is given, the element becomes a member of the global
    # namespace.
    #
    def add_element(self, namespace, name, elem):

        if (namespace): self.__make_namespace(namespace)

        wrapped = ElementWrapper(elem)
        if (not namespace):
            self.__environment[name] = wrapped
        else:
            setattr(self.__environment[namespace], name, wrapped)


    #
    # Same as add_element() but puts elements into a structure of (nested)
    # arrays.
    #
    def add_element_with_path(self, namespace, name, elem, indexpath):

        if (namespace): self.__make_namespace(namespace)

        try:
            if (not namespace):
                lst = self.__environment[name]
            else:
                lst = getattr(self.__environment[namespace], name)
        except:
            lst = []

        if (not namespace):
            self.__environment[name] = lst
        else:
            setattr(self.__environment[namespace], name, lst)

        # build up the structure of (nested) arrays
        for index in indexpath[:-1]:
            while (len(lst) - 1 < index): lst.append([])
            lst = lst[index]
        index = indexpath[-1]
        while (len(lst) - 1 < index): lst.append([])
        lst[index] = ElementWrapper(elem)



    #
    # Fixes the indentation of Python code.
    #
    def __fix_indentation(self, code):

        lines = code.splitlines()
        min_indent = len(code)
        # find the minimal indentation
        for l in lines:
            if (not l.strip()): continue
            this_indent = len(l) - len(l.lstrip())
            min_indent = min(min_indent, this_indent)

        # apply the minimal indentation
        out = ""
        for l in lines:
            out += l[min_indent:] + "\n"

        return out



    #
    # Executes a block of script.
    #
    def execute(self, scriptlet, handle_error = True):

        sid = scriptlet.script_id

        # get the block into shape
        code = self.__fix_indentation(scriptlet.script)

        # remember scriptlet for later
        self.__scriptlets[sid] = scriptlet

        # compile and run
        try:
            from utils.error import Error
            Error().register_code("<inline '%s'>" % sid, code)
            pycode = compile(code, "<inline '%s'>" % sid, 'exec')
            #pycode = compile(code, "%s" % scriptlet.filename, 'exec')
            exec pycode in self.__environment

        except:
            #if (handle_error):
            self.__handle_error()



    #
    # Retrieves the value of the given object from the scripting environment.
    #
    def get_value(self, name):

        if (not name):
            return
        # TODO: if the type of the bound variable differs we have to check it
        cmd = "__retrieve__ = %s" % (name,)
        self.execute(Scriptlet(cmd, "<internal>"), handle_error = False)

        # may raise an exception
        return self.__environment.pop("__retrieve__")



    #
    # Sets the value of the given object in the scripting environment.
    #
    def set_value(self, name, value):

        self.__environment["__inject__"] = value
        cmd = "%s = __inject__" % name
        self.execute(Scriptlet(cmd, "<internal>"), handle_error = False)
        self.__environment.pop("__inject__")



    #
    # Calls the given function in the sandbox.
    #
    def call_function(self, name, *args):

        func = self.get_value(name)
        try:
            func(*args)
        except:
            #log("A function call in the inline script failed.")
            self.__handle_error()



    #BEGIN
    # + 2.3.3 http://python.org/doc/2.3.3/lib/built-in-funcs.html
    #                        vs.
    # - 2.2.3 http://python.org/doc/2.2.3/lib/built-in-funcs.html
    # * dangerous

    # * __import__
    # abs
    # - apply
    # + basestring
    # bool
    # - buffer
    # callable
    # chr
    # classmethod
    # cmp
    # - coerce
    # * compile
    # complex
    # delattr
    # dict
    # dir # ?
    # divmod
    # + enumerate
    # * eval
    # * execfile
    # * file
    # filter # may be should print a warning and a link to list-comprehension
    # float
    # getattr
    # globals # ?
    # hasattr
    # hash
    # help # not very useful
    # hex
    # id
    # - intern
    # input # not very useful
    # int
    # isinstance
    # issubclass
    # iter
    # len
    # list
    # locals
    # long
    # map # maybe should print a warning and a link to list-comprehension
    # max
    # min
    # + object
    # oct
    # * open
    # ord
    # property
    # range
    # raw_input # not very usefull
    # reduce
    # * reload
    # repr
    # round
    # setattr
    # - slice
    # staticmethod
    # str
    # + sum
    # super
    # tuple
    # type
    # unichr
    # unicode
    # vars
    # xrange
    # zip

    # 2.3 Deprecated/Non-essentials functions
    # http://python.org/doc/2.3.3/lib/non-essential-built-in-funcs.html
    # - apply
    # - buffer
    # - coerce
    # - intern
    #END
